"""
Package which defines modules that facilitate RVC based audio
generation.
"""

from __future__ import annotations

import static_ffmpeg
import static_sox

static_ffmpeg.add_paths()
static_sox.add_paths()
